"""Handle login and downloading from the USGS M2M portal."""

import os
import re

import requests
from tqdm import tqdm
import random
import string
import threading
from usgsm2m.api import API
from usgsm2m.errors import USGSM2MError
from usgsm2m.util import guess_dataset, is_display_id


def _get_tokens(body):
    """Get `csrf_token` and `__ncforminfo`."""
    csrf = re.findall(r'name="csrf" value="(.+?)"', body)[0]
    ncform = re.findall(r'name="__ncforminfo" value="(.+?)"', body)[0]

    if not csrf:
        raise USGSM2MError("USGSM2M: login failed (csrf token not found).")
    if not ncform:
        raise USGSM2MError("USGSM2M: login failed (ncforminfo not found).")

    return csrf, ncform


class USGSM2M(object):
    """Access USGSM2M portal."""

    def __init__(self, username, password):
        """Access USGSM2M portal."""
        self.session = requests.Session()
        #self.login(username, password)
        self.api = API(username, password)
        self.maxthreads = 5 # Threads count for downloads
        self.sema = threading.Semaphore(value=self.maxthreads)
        self.threads = []

    def logged_in(self):
        """Check if the log-in has been successfull based on session cookies."""
        eros_sso = self.session.cookies.get("EROS_SSO_production_secure")
        return bool(eros_sso)

    def login(self, username, password):
        """Login to USGSM2M."""
        rsp = self.session.get(EE_LOGIN_URL)
        csrf, ncform = _get_tokens(rsp.text)
        payload = {
            "username": username,
            "password": password,
            "csrf": csrf,
            "__ncforminfo": ncform,
        }
        rsp = self.session.post(EE_LOGIN_URL, data=payload, allow_redirects=True)

        if not self.logged_in():
            raise USGSM2MError("EE: login failed.")

    def logout(self):
        """Log out from USGSM2M."""
        self.api.logout()

    def _download(self, url, output_dir, timeout, chunk_size=1024, skip=False):
        """Download remote file given its URL."""
        # Check availability of the requested product
        # USGSM2M should respond with JSON
        with self.session.get(
            url, allow_redirects=False, stream=True, timeout=timeout
        ) as r:
            r.raise_for_status()
            error_msg = r.json().get("errorMessage")
            if error_msg:
                raise USGSM2MError(error_msg)
            download_url = r.json().get("url")

        try:
            with self.session.get(
                download_url, stream=True, allow_redirects=True, timeout=timeout
            ) as r:
                file_size = int(r.headers.get("Content-Length"))
                with tqdm(
                    total=file_size, unit_scale=True, unit="B", unit_divisor=1024
                ) as pbar:
                    local_filename = r.headers["Content-Disposition"].split("=")[-1]
                    local_filename = local_filename.replace('"', "")
                    local_filename = os.path.join(output_dir, local_filename)
                    if skip:
                        return local_filename
                    with open(local_filename, "wb") as f:
                        for chunk in r.iter_content(chunk_size=chunk_size):
                            if chunk:
                                f.write(chunk)
                                pbar.update(chunk_size)
        except requests.exceptions.Timeout:
            raise USGSM2MError(
                "Connection timeout after {} seconds.".format(timeout)
            )
        return local_filename

    def download(self, identifier, output_dir,filetype='bundle', idField='displayId', dataset=None, timeout=300,chunk_size=1024, skip=False):
        """Download a Landsat scene.

        Parameters
        ----------
        identifier : str
            Scene Entity ID or Display ID.
        output_dir : str
            Output directory. Automatically created if it does not exist.
        dataset : str, optional
            Dataset name. If not provided, automatically guessed from scene id.
        timeout : int, optional
            Connection timeout in seconds.
        skip : bool, optional
            Skip download, only returns the remote filename.

        Returns
        -------
        filename : str
            Path to downloaded file.
        """
        #print("usgsm2m.py:download",identifier)
        productsdownloads=self.api.get_products_download_options(entityList=identifier,filetype=filetype,datasetName=dataset,idField=idField)
        #print(productsdownloads)
        productsUrls=self.api.get_download_urls(productsdownloads)
        if skip :
            return productsUrls
        #print(productsUrls)
        for productsUrl in productsUrls:
            self.runDownloadMultiThread(self.threads, output_dir,timeout,chunk_size,skip,productsUrl) 


    def resume_download(self,output_dir,timeout,chunk_size,skip,download_url):
        self.sema.acquire() 
        try:
            with self.session.get(
                download_url, stream=True, allow_redirects=True, timeout=timeout
            ) as rh:
                file_size = int(rh.headers.get("Content-Length",0))
                local_filename = rh.headers["Content-Disposition"].split("=")[-1]
                local_filename = local_filename.replace('"', "")
                if local_filename.endswith('tar'):
                    local_filename=local_filename+".gz"
                local_filename = os.path.join(output_dir, local_filename)
                resume_byte_pos=self.checkFileSize(local_filename)
                initial_pos = resume_byte_pos if resume_byte_pos else 0
                mode = 'ab' if resume_byte_pos else 'wb' 
                resume_header = ({'Range': f'bytes={resume_byte_pos}-'} if resume_byte_pos else None)
                with self.session.get(
                    download_url, stream=True, allow_redirects=True, timeout=timeout, headers=resume_header
                ) as r:
                    with open(local_filename, mode) as f:
                        with tqdm(total=file_size, unit='B',unit_scale=True, unit_divisor=1024,desc=local_filename, 
                                    initial=initial_pos, miniters=1) as pbar:
                            for chunk in r.iter_content(chunk_size=chunk_size):
                                if chunk:
                                    f.write(chunk)
                                    pbar.update(chunk_size)
            self.sema.release()
        except requests.exceptions.Timeout:
            raise USGSM2MError(
                "Connection timeout after {} seconds.".format(timeout)
            )
            self.sema.release()
            self.runDownloadMultiThread(threads, output_dir, timeout, chunk_size, skip, download_url)
        return local_filename

    def checkFileSize(self,local_filename):
        if os.path.isfile(local_filename):
            return os.stat(local_filename).st_size
        return None

    def _downloadFileMultiThread(self,output_dir,timeout,chunk_size,skip,download_url):
        self.sema.acquire()
        local_filename=None
        try:
            with self.session.get(
                download_url, stream=True, allow_redirects=True, timeout=timeout
            ) as r:
                file_size = int(r.headers.get("Content-Length"))
                with tqdm(
                    total=file_size, unit_scale=True, unit="B", unit_divisor=1024
                ) as pbar:
                    local_filename = r.headers["Content-Disposition"].split("=")[-1]
                    local_filename = local_filename.replace('"', "")+".gz"
                    local_filename = os.path.join(output_dir, local_filename)
                    if skip:
                        return local_filename
                    with open(local_filename, "wb") as f:
                        for chunk in r.iter_content(chunk_size=chunk_size):
                            if chunk:
                                f.write(chunk)
                                pbar.update(chunk_size)
            #self.threads.remove(threading.currentThread())
            self.sema.release()
        except requests.exceptions.Timeout:
            raise USGSM2MError(
                "Connection timeout after {} seconds.".format(timeout)
            )
            self.sema.release()

            self.runDownloadMultiThread(threads, output_dir, timeout, chunk_size, skip, download_url)
        return local_filename
  
    def runDownloadMultiThread(self,threads, output_dir,timeout,chunk_size,skip,url):
        #thread = threading.Thread(target=self._downloadFileMultiThread, args=(output_dir,timeout,chunk_size,skip,url,))
        thread = threading.Thread(target=self.resume_download, args=(output_dir,timeout,chunk_size,skip,url,))
        self.threads.append(thread)
        thread.start()

    def downloadbulk(self,entityfile, output_dir,dataset=None, filetype="bundle",idField="displayId",timeout=300, chunk_size=1024,skip=False):
        """Download a Landsat scene.

        Parameters
        ----------
        entityfile : str
            Scene Entity ID or Display ID File Containing their entries.
        dataset : str, optional
                    Dataset name. If not provided, automatically guessed from scene id.
        output_dir : str
            Output directory. Automatically created if it does not exist.
        filetype : str 
            File type as 'bundle' or 'bandd' for individual files 
        idFiled : str
            Files contains the 'entityId' or 'displayId'
        timeout : int, optional
            Connection timeout in seconds.
        skip : bool, optional
            Skip download, only returns the remote filename list.

        Returns
        -------
        filename : str
            Path to downloaded file.
        """

        try:
            with open(entityfile, 'r') as f:
                entityList = f.readlines()   
            if(len(entityList)<1):
                raise USGSM2MError(
                    "Entity File:{entityfile} is Empty.".format(entityfile)
                )
        except FileNotFoundError as e:
            raise USGSM2MError(
                "Entity File:{entityfile} not found.".format(entityfile)
            )
        productsdownloads=self.api.get_products_download_options(entityList=entityList,filetype=filetype,datasetName=dataset,idField=idField)
        #print(productsdownloads)
        productsUrls=self.api.get_download_urls(productsdownloads)
        if skip :
            return productsUrls
        #print(productsUrls)
        for productsUrl in productsUrls:
            self.runDownloadMultiThread(self.threads, output_dir,timeout,chunk_size,skip,productsUrl) 

        print("Download Started\n")
        for thread in self.threads:
            thread.join()





"""Utility functions."""

from usgsm2m.errors import USGSM2MError


def _is_landsat_product_id(id):
    return len(id) == 40 and id.startswith("L")


def _is_landsat_scene_id(id):
    return len(id) == 21 and id.startswith("L")


def _is_sentinel_display_id(id):
    return len(id) == 34 and id.startswith("L")


def _is_sentinel_entity_id(id):
    return len(id) == 8 and id.isdecimal()


def is_display_id(id):
    return _is_landsat_product_id(id) or _is_sentinel_display_id(id)


def is_entity_id(id):
    return _is_landsat_scene_id(id) or _is_sentinel_entity_id(id)


def is_product_id(identifier):
    """Check if a given identifier is a product identifier
    as opposed to a legacy scene identifier.
    """
    return len(identifier) == 40 and identifier.startswith("L")


def parse_product_id(product_id):
    """Retrieve information from a product identifier.

    Parameters
    ----------
    product_id : str
        Landsat product identifier (also referred as Display ID).

    Returns
    -------
    meta : dict
        Retrieved information.
    """
    elements = product_id.split("_")
    return {
        "product_id": product_id,
        "sensor": elements[0][1],
        "satellite": elements[0][2:4],
        "processing_level": elements[1],
        "satellite_orbits": elements[2],
        "acquisition_date": elements[3],
        "processing_date": elements[4],
        "collection_number": elements[5],
        "collection_category": elements[6],
    }


def parse_scene_id(scene_id):
    """Retrieve information from a scene identifier.

    Parameters
    ----------
    scene_id : str
        Landsat scene identifier (also referred as Entity ID).

    Returns
    -------
    meta : dict
        Retrieved information.
    """
    return {
        "scene_id": scene_id,
        "sensor": scene_id[1],
        "satellite": scene_id[2],
        "path": scene_id[3:6],
        "row": scene_id[6:9],
        "year": scene_id[9:13],
        "julian_day": scene_id[13:16],
        "ground_station": scene_id[16:19],
        "archive_version": scene_id[19:21],
    }


def landsat_dataset(satellite, collection="c1", level="l1"):
    """Get landsat dataset name."""
    if satellite == 5:
        sensor = "tm"
    elif satellite == 7:
        sensor = "etm"
    elif satellite == 8 and collection == "c1":
        sensor = "8"
    elif satellite == 8 and collection == "c2":
        sensor = "ot"
    else:
        raise USGSM2MError("Failed to guess dataset from identifier.")
    dataset = f"landsat_{sensor}_{collection}"
    if collection == "c2":
        dataset += f"_{level}"
    return dataset


def guess_dataset(identifier):
    """Guess data set based on a scene identifier."""
    # Landsat Product Identifier
    #print("util.py:guess_dataset"+identifier)
    if _is_landsat_product_id(identifier):
        meta = parse_product_id(identifier)
        satellite = int(meta["satellite"])
        collection = "c" + meta["collection_number"][-1]
        level = meta["processing_level"][:2].lower()
        return landsat_dataset(satellite, collection, level)
    elif _is_landsat_scene_id(identifier):
        meta = parse_scene_id(identifier)
        satellite = int(meta["satellite"])
        return landsat_dataset(satellite)
    elif _is_sentinel_display_id(identifier) or _is_sentinel_entity_id(identifier):
        return "sentinel_2a"
    else:
        raise USGSM2MError("Failed to guess dataset from identifier.")


def title_to_snake(src_string):
    """Convert title string to snake_case."""
    return src_string.lower().replace(" ", "_").replace("/", "-")


def camel_to_snake(src_string):
    """Convert camelCase string to snake_case."""
    dst_string = [src_string[0].lower()]
    for c in src_string[1:]:
        if c in ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"):
            dst_string.append("_")
            dst_string.append(c.lower())
        else:
            dst_string.append(c)
    return "".join(dst_string)



class USGSError(Exception):
    """General error with USGS API."""

    pass


class USGSInvalidEndpoint(Exception):
    """Endpoint is invalid."""

    pass


class USGSInvalidParametersError(Exception):
    """Provided parameters are invalid."""

    pass


class USGSUnauthorizedError(Exception):
    """User does not have access to the requested endpoint."""

    pass


class USGSAuthenticationError(Exception):
    """User credentials verification failed or API key is invalid."""

    pass


class USGSRateLimitError(Exception):
    """Account does not support multiple requests at a time."""

    pass


class USGSM2MError(Exception): 
    """Exception for errors raised by the USGSM2 API."""

    pass




"""Python implementation of the USGSM2M API."""

import json
from urllib.parse import urljoin
import string
import random
from datetime import datetime
from dateutil import parser
import time

import requests
from shapely.geometry import Point, shape
from usgsm2m.util import is_display_id,is_entity_id
from usgsm2m.errors import USGSAuthenticationError, USGSError, USGSRateLimitError
from usgsm2m.util import guess_dataset

API_URL = "https://m2m.cr.usgs.gov/api/api/json/stable/"


class API(object):
    """USGSM2M API."""

    def __init__(self, username, password):
        """USGSM2M API.

        Parameters
        ----------
        username : str
            USGS USGSM2M username.
        password : str
            USGS USGSM2M password.
        """
        self.url = API_URL
        self.session = requests.Session()
        self.login(username, password)

    @staticmethod
    def raise_api_error(response):
        """Parse API response and return the appropriate exception.

        Parameters
        ----------
        response : requests response
            Response from USGS API.
        """
        data = response.json()
        error_code = data.get("errorCode")
        error_msg = data.get("errorMessage")
        if error_code:
            if error_code in ("AUTH_INVALID", "AUTH_UNAUTHROIZED", "AUTH_KEY_INVALID"):
                raise USGSAuthenticationError(f"{error_code}: {error_msg}.")
            elif error_code == "RATE_LIMIT":
                raise USGSRateLimitError(f"{error_code}: {error_msg}.")
            else:
                raise USGSError(f"{error_code}: {error_msg}.")

    def request(self, endpoint, params=None):
        """Perform a request to the USGS M2M API.

        Parameters
        ----------
        endpoint : str
            API endpoint.
        params : dict, optional
            API parameters.

        Returns
        -------
        data : dict
            JSON data returned by the USGS API.

        Raises
        ------
        USGSAuthenticationError
            If credentials are not valid of if user lacks permission.
        USGSError
            If the USGS API returns a non-null error code.
        """
        url = urljoin(self.url, endpoint)
        data = json.dumps(params)
        #print("apt.py:request",data)
        r = self.session.get(url, data=data)
        try:
            self.raise_api_error(r)
        except USGSRateLimitError:
            time.sleep(3)
            r = self.session.get(url, data=data)
        self.raise_api_error(r)
        return r.json().get("data")

    def login(self, username, password):
        """Get an API key.

        Parameters
        ----------
        username : str
            USGSM2M username.
        password : str
            USGSM2M password.
        """
        login_url = urljoin(self.url, "login")
        payload = {"username": username, "password": password}
        r = self.session.post(login_url, json.dumps(payload))
        self.raise_api_error(r)
        self.session.headers["X-Auth-Token"] = r.json().get("data")

    def logout(self):
        """Logout from USGS M2M API."""
        self.request("logout")
        self.session = requests.Session()

    def get_entity_id(self, display_id, dataset):
        """Get scene ID from product ID.

        Note
        ----
        As the lookup endpoint has been removed in API v1.5, the function makes
        successive calls to scene-list-add and scene-list-get in order to retrieve
        the scene IDs. A temporary sceneList is created and removed at the end of the
        process.

        Parameters
        ----------
        display_id : str or list of str
            Input product ID. Can also be a list of product IDs.
        dataset : str
            Dataset alias.

        Returns
        -------
        entity_id : str or list of str
            Output entity ID. Can also be a list of entity IDs depending on input.
        """
        # scene-list-add support both entityId and entityIds input parameters
        param = "entityId"
        if isinstance(display_id, list):
            param = "entityIds"

        # a random scene list name is created -- better error handling is needed
        # to ensure that the temporary scene list is removed even if scene-list-get
        # fails.
        list_id = _random_string()
        self.request(
            "scene-list-add",
            params={
                "listId": list_id,
                "datasetName": dataset,
                "idField": "displayId",
                param: display_id,
            },
        )
        r = self.request("scene-list-get", params={"listId": list_id})
        entity_id = [scene["entityId"] for scene in r]
        self.request("scene-list-remove", params={"listId": list_id})

        if param == "entityId":
            return entity_id[0]
        else:
            return entity_id

    def metadata(self, entity_id, dataset, browse=False):
        """Get metadata for a given scene.

        Parameters
        ----------
        entity_id : str
            Landsat Scene ID or Sentinel Entity ID.
        dataset : str
            Dataset alias.
        browse : bool, optional
            Include browse (LandsatLook URLs) metadata items.

        Returns
        -------
        meta : dict
            Scene metadata.
        """
        r = self.request(
            "scene-metadata",
            params={
                "datasetName": dataset,
                "entityId": entity_id,
                "metadataType": "full",
            },
        )
        return _parse_metadata(r, parse_browse_field=browse)

    def get_display_id(self, entity_id, dataset):
        """Get display ID from entity ID.

        Parameters
        ----------
        entity_id : str
            LLandsat Scene ID or Sentinel Entity ID.
        dataset : str
            Dataset alias.

        Returns
        -------
        display_id : str
            Landsat Product ID or Sentinel Display ID.
        """
        meta = self.metadata(entity_id, dataset)
        return meta["display_id"]

    def search(
        self,
        dataset,
        longitude=None,
        latitude=None,
        bbox=None,
        max_cloud_cover=None,
        start_date=None,
        end_date=None,
        months=None,
        max_results=100,
    ):
        """Search for scenes.

        Parameters
        ----------
        dataset : str
            Case-insensitive dataset alias (e.g. landsat_tm_c1).
        longitude : float, optional
            Longitude of the point of interest.
        latitude : float, optional
            Latitude of the point of interest.
        bbox : tuple, optional
            (xmin, ymin, xmax, ymax) of the bounding box.
        max_cloud_cover : int, optional
            Max. cloud cover in percent (1-100).
        start_date : str, optional
            YYYY-MM-DD
        end_date : str, optional
            YYYY-MM-DD. Equal to start_date if not provided.
        months : list of int, optional
            Limit results to specific months (1-12).
        max_results : int, optional
            Max. number of results. Defaults to 100.

        Returns
        -------
        scenes : list of dict
            Matching scenes as a list of dict containing metadata.
        """
        spatial_filter = None
        if longitude and latitude:
            spatial_filter = SpatialFilterMbr(*Point(longitude, latitude).bounds)
        elif bbox:
            spatial_filter = SpatialFilterMbr(*bbox)

        acquisition_filter = None
        if start_date and end_date:
            acquisition_filter = AcquisitionFilter(start_date, end_date)

        cloud_cover_filter = None
        if max_cloud_cover:
            cloud_cover_filter = CloudCoverFilter(
                max=max_cloud_cover, include_unknown=False
            )

        scene_filter = SceneFilter(
            acquisition_filter, spatial_filter, cloud_cover_filter, months=months
        )

        r = self.request(
            "scene-search",
            params={
                "datasetName": dataset,
                "sceneFilter": scene_filter,
                "maxResults": max_results,
                "metadataType": "full",
            },
        )
        return [_parse_metadata(scene) for scene in r.get("results")]

    def get_products_download_options(self,entityList,filetype,datasetName=None,idField="displayId"):
        entityIdsinFile = {}
        finalDownload = []
        # Add scenes to a list
        listname=_random_string()

        for entity in entityList:
            entitykey=entity.strip().encode("ascii", "ignore").decode()
            #If datasetName is not given guess it
            #print("api.py:get_products_download_options",entity)

            if not datasetName :
                datasetName=guess_dataset(entitykey) 

            if is_display_id(entitykey):
               idField="displayId" 
            elif is_entity_id(entitykey):
               idField="entityId"
            else:
                raise USGSError(
                "Entity is neither displayId or entityId :{} not found.".format(entitykey)) 

            listId = f"temp_{datasetName}_{listname}_list" # customized list id
            payload = {
                "listId": listId,
                "idField" : idField,
                "datasetName": datasetName
            }

            payloadkey=json.dumps(payload)
            if payloadkey in entityIdsinFile.keys() :
                entityIdsinFile[payloadkey].append(entitykey) 
            else :
                entityIdsinFile[payloadkey]=[entitykey]
            

        for payloadkey in entityIdsinFile.keys(): 
            
            payload=json.loads(payloadkey)
            payload["entityIds"]=entityIdsinFile[payloadkey]

            self.request("scene-list-add",params=payload)

            # Get download options
            payload = {
                "listId": listId,
                "datasetName": datasetName
            }

            products=self.request("download-options",params=payload)

           # Select products
            downloads = []
            if filetype == 'bundle':
                # select bundle files
                for product in products:        
                    if product["bulkAvailable"]:               
                        downloads.append({"entityId":product["entityId"], "productId":product["id"]})
            elif filetype == 'band':
                # select band files
                for product in products:  
                    if product["secondaryDownloads"] is not None and len(product["secondaryDownloads"]) > 0:
                        for secondaryDownload in product["secondaryDownloads"]:
                            if secondaryDownload["bulkAvailable"]:
                                downloads.append({"entityId":secondaryDownload["entityId"], "productId":secondaryDownload["id"]})
            else:
                # select all available files
                for product in products:        
                    if product["bulkAvailable"]:               
                        downloads.append({"entityId":product["entityId"], "productId":product["id"]})
                    if product["secondaryDownloads"] is not None and len(product["secondaryDownloads"]) > 0:
                        for secondaryDownload in product["secondaryDownloads"]:
                            if secondaryDownload["bulkAvailable"]:
                                downloads.append({"entityId":secondaryDownload["entityId"], "productId":secondaryDownload["id"]})
            # Remove the list
            payload = {
                "listId": listId
            }
            self.request("scene-list-remove",params=payload)
            finalDownload=finalDownload+downloads

        return finalDownload
          
    def get_download_urls(self,downloads):
            # Send download-request
        productURLs=[]
        label = datetime.now().strftime("%Y%m%d_%H%M%S") # Customized label using date time
        payload = {
            "downloads": downloads,
            "label": label,
            'returnAvailable': True
        }
        
        #print(f"Sending download request ...\n")
        results = self.request("download-request", params=payload)
        #print(f"Done sending download request\n") 

          
        for result in results['availableDownloads']:       
            #print(f"Get download url: {result['url']}\n" )
            productURLs.append(result['url'])
            #runDownload(threads, result['url'])

        preparingDownloadCount = len(results['preparingDownloads'])
        preparingDownloadIds = []
        if preparingDownloadCount > 0:
            for result in results['preparingDownloads']:  
                preparingDownloadIds.append(result['downloadId'])
      
            payload = {"label" : label}                
            # Retrieve download urls
            #print("Retrieving download urls...\n")
            results = self.request("download-retrieve", params=payload)
            if results != False:
                for result in results['available']:
                    if result['downloadId'] in preparingDownloadIds:
                        preparingDownloadIds.remove(result['downloadId'])
                        #print(f"Get download url: {result['url']}\n" )
                        productURLs.append(result['url'])
                        #runDownload(threads, result['url'])
                    
                for result in results['requested']:   
                    if result['downloadId'] in preparingDownloadIds:
                        preparingDownloadIds.remove(result['downloadId'])
                        #print(f"Get download url: {result['url']}\n" )
                        productURLs.append(result['url'])
                        #runDownload(threads, result['url'])
            
            # Don't get all download urls, retrieve again after 30 seconds
            while len(preparingDownloadIds) > 0: 
                print(f"{len(preparingDownloadIds)} downloads are not available yet. Waiting for 30s to retrieve again\n")
                time.sleep(30)
                results =  self.request("download-retrieve",params=payload)
                if results != False:
                    for result in results['available']:                            
                        if result['downloadId'] in preparingDownloadIds:
                            preparingDownloadIds.remove(result['downloadId'])
                            #print(f"Get download url: {result['url']}\n" )
                            productURLs.append(result['url'])
                            #runDownload(threads, result['url'])
        return productURLs

def _random_string(length=10):
    """Generate a random string."""
    letters = string.ascii_lowercase
    return "".join(random.choice(letters) for i in range(length))


def _title_to_snake(src_string):
    """Convert title case to snake_case."""
    return src_string.lower().replace(" ", "_").replace("/", "-")


def _camel_to_snake(src_string):
    """Convert camelCase string to snake_case."""
    dst_string = [src_string[0].lower()]
    for c in src_string[1:]:
        if c in ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"):
            dst_string.append("_")
            dst_string.append(c.lower())
        else:
            dst_string.append(c)
    return "".join(dst_string)


def _to_num(src_string):
    """Convert string to int or float if possible.

    Original value is returned if conversion failed.
    """
    if not isinstance(src_string, str):
        return src_string
    src_string = src_string.strip()
    try:
        return int(src_string)
    except ValueError:
        try:
            return float(src_string)
        except ValueError:
            return src_string


def _to_date(src_string):
    """Convert string to datetime if possible.

    Original value is returned if conversion failed.
    """
    if not isinstance(src_string, str):
        return src_string
    try:
        return parser.parse(src_string)
    except parser.ParserError:
        try:
            # Specific date format for start_time and end_time
            nofrag, frag = src_string.split(".")
            dtime = datetime.strptime(nofrag, "%Y:%j:%H:%M:%S")
            dtime = dtime.replace(microsecond=int(frag[:6]))
            return dtime
        except ValueError:
            pass
    return src_string


def _parse_value(src_value):
    """Try to convert value to numeric or date if possible.

    Original value is returned if conversion failed.
    """
    dst_value = src_value
    if isinstance(dst_value, str):
        dst_value = dst_value.strip()
        dst_value = _to_num(dst_value)
        dst_value = _to_date(dst_value)
    return dst_value


def _parse_browse_metadata(src_meta):
    """Parse the browse field returned by the API."""
    dst_meta = {}
    for product in src_meta:
        name = _title_to_snake(product["browseName"])
        dst_meta[name] = {}
        for field, value in product.items():
            dst_meta[name][_camel_to_snake(field)] = value
    return dst_meta


def _parse_metadata_field(src_meta):
    """Parse the metadata field returned by the API."""
    dst_meta = {}
    for meta in src_meta:
        # Convert field name to snake case
        name = _title_to_snake(meta["fieldName"])
        # Abbreviate "identifier" by "id" for shorter names
        name = name.replace("identifier", "id")
        # Always use "acquisition_date" instead of "acquired_date" for consistency
        if name == "date_acquired":
            name = "acquisition_date"
        # Remove processing-level information in field names for consistency
        name = name.replace("_l1", "")
        name = name.replace("_l2", "")
        # Dictionary link URL also provides some information on the field
        dict_id = meta.get("dictionaryLink").split("#")[-1].strip()
        # Do not process this field
        if dict_id == "coordinates_degrees":
            continue
        # Sentinel metadata has an "Entity ID" field that would
        # conflict with the API entityId field
        if name == "entity_id":
            name = "sentinel_entity_id"
        # Do not parse numeric IDs. Keep them as strings.
        if name.endswith("_id"):
            dst_meta[name] = str(meta.get("value")).strip()
        else:
            dst_meta[name] = _parse_value(meta.get("value"))
    return dst_meta


def _parse_metadata(response, parse_browse_field=False):
    """Parse the full response returned by the API when requesting metadata."""
    metadata = {}
    for key, value in response.items():
        name = _camel_to_snake(key)
        if key == "browse":
            if parse_browse_field:
                metadata[name] = _parse_browse_metadata(value)
            else:
                continue
        elif key == "spatialCoverage":
            metadata[name] = shape(value)
        elif key == "spatialBounds":
            metadata[name] = shape(value).bounds
        elif key == "temporalCoverage":
            start, end = value["endDate"], value["startDate"]
            metadata[name] = [_to_date(start), _to_date(end)]
        elif key == "metadata":
            metadata.update(_parse_metadata_field(value))
        else:
            # Do not parse numeric IDs. Keep them as strings.
            if name.endswith("_id"):
                metadata[name] = str(value).strip()
            else:
                metadata[name] = _parse_value(value)
    if "acquisition_date" not in metadata:
        metadata["acquisition_date"] = metadata["temporal_coverage"][0]
    return metadata


class Coordinate(dict):
    """A coordinate object as expected by the USGS M2M API.

    Parameters
    ----------
    longitude : float
        Decimal longitude.
    latitude : float
        Decimal latitude.
    """

    def __init__(self, longitude, latitude):
        self["longitude"] = longitude
        self["latitude"] = latitude


class GeoJson(dict):
    """A GeoJSON object as expected by the USGS M2M API.

    Parameters
    ----------
    shape : dict
        Input geometry as a geojson-like dict.
    """

    def __init__(self, shape):
        self["type"] = shape["type"]
        self["coordinates"] = self.transform(shape["type"], shape["coordinates"])

    @staticmethod
    def transform(type, coordinates):
        """Convert geojson-like coordinates as expected by the USGS M2M API.

        Essentially converts tuples of coordinates to api.Coordinate objects.
        """
        if type == "MultiPolygon":
            return [
                [Coordinate(*point) for point in polygon] for polygon in coordinates[0]
            ]
        elif type == "Polygon":
            return [Coordinate(*point) for point in coordinates[0]]
        elif type == "LineString":
            return [Coordinate(*point) for point in coordinates]
        elif type == "Point":
            return Coordinate(*coordinates)
        else:
            raise ValueError(f"Geometry type `{type}` not supported.")


class SpatialFilterMbr(dict):
    """Bounding box spatial filter.

    Parameters
    ----------
    xmin : float
        Min. decimal longitude.
    ymin : float
        Min. decimal latitude.
    xmax : float
        Max. decimal longitude.
    ymax : float
        Max. decimal latitude.
    """

    def __init__(self, xmin, ymin, xmax, ymax):
        self["filterType"] = "mbr"
        self["lowerLeft"] = Coordinate(xmin, ymin)
        self["upperRight"] = Coordinate(xmax, ymax)


class SpatialFilterGeoJSON(dict):
    """GeoJSON-based spatial filter.

    Parameters
    ----------
    shape : dict
        Input shape as a geojson-like dict.
    """

    def __init__(self, shape):
        self["filterType"] = "geoJson"
        self["geoJson"] = GeoJson(shape)


class AcquisitionFilter(dict):
    """Acquisition date filter.

    Parameters
    ----------
    start : str
        ISO 8601 start date.
    end : str
        ISO 8601 end date.
    """

    def __init__(self, start, end):
        self["start"] = start
        self["end"] = end


class CloudCoverFilter(dict):
    """Cloud cover filter.

    Parameters
    ----------
    min : int, optional
        Min. cloud cover in percents (default=0).
    max : int, optional
        Max. cloud cover in percents (default=100).
    include_unknown : bool, optional
        Include scenes with unknown cloud cover (default=False).
    """

    def __init__(self, min=0, max=100, include_unknown=False):
        self["min"] = min
        self["max"] = max
        self["includeUnknown"] = include_unknown


class MetadataValue(dict):
    """Metadata filter.

    Parameters
    ----------
    field_id : str
        ID of the field.
    value : str, float or int
        Value of the field.
    """

    def __init__(self, field_id, value):
        self["filterType"] = "value"
        self["filterId"] = field_id
        self["value"] = value
        if isinstance(value, str):
            self["operand"] = "like"
        else:
            self["operand"] = "="


class SceneFilter(dict):
    """Scene search filter.

    Parameters
    ----------
    acquisition_filter : AcquisitionFilter, optional
        Acquisition date filter.
    spatial_filter : SpatialFilterMbr or SpatialFilterGeoJson, optional
        Spatial filter.
    cloud_cover_filter : CloudCoverFilter, optional
        Cloud cover filter.
    metadata_filter : MetadataValue, optional
        Metadata filter.
    months : list of int, optional
        Seasonal filter (month numbers from 1 to 12).
    """

    def __init__(
        self,
        acquisition_filter=None,
        spatial_filter=None,
        cloud_cover_filter=None,
        metadata_filter=None,
        months=None,
    ):
        if acquisition_filter:
            self["acquisitionFilter"] = acquisition_filter
        if spatial_filter:
            self["spatialFilter"] = spatial_filter
        if cloud_cover_filter:
            self["cloudCoverFilter"] = cloud_cover_filter
        if metadata_filter:
            self["metadataFilter"] = metadata_filter
        if months:
            self["seasonalFilter"] = months



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['usgsm2m']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1,<8.0',
 'python-dateutil>=2.8,<3.0',
 'requests>=2.20,<3.0',
 'shapely>=1.7,<2.0',
 'tqdm>=4.58,<5.0']

entry_points = \
{'console_scripts': ['usgsm2m = usgsm2m.cli:cli']}

setup_kwargs = {
    'name': 'usgsm2m',
    'version': '0.10.0',
    'description': 'Search and download Landsat scenes from USGSM2M',
    'long_description': '[![Build](https://img.shields.io/github/workflow/status/ashutoshkumarjha/usgsm2m/Upload%20Python%20Package?label=build&logo=github)](https://github.com/ashutoshkumarjha/usgsm2m/actions/workflows/python-publish.yml)\n[![Tests](https://img.shields.io/github/workflow/status/ashutoshkumarjha/usgsm2m/Run%20tests?label=tests&logo=github)](https://github.com/ashutoshkumarjha/usgsm2m/actions/workflows/run-tests.yml)\n[![codecov](https://codecov.io/gh/ashutoshkumarjha/usgsm2m/branch/master/graph/badge.svg?token=NwVo09Edur)](https://codecov.io/gh/ashutoshkumarjha/usgsm2m)\n\n\n# Description\n\n![CLI Demo](https://raw.githubusercontent.com/ashutoshkumarjha/usgsm2m/master/demo.gif?s=0.5)\n\nThe **usgsm2m** Python package provides an interface to the [USGS M2M](https://m2m.cr.usgs.gov) portal to search and download scenes through a command-line interface or a Python API.\n\nThe following datasets are supported based on only entityid provided. For the different dataset other than this list requires the dataset name. The bulk download is also supported with resume facility.\n\n\n| Dataset Name | Dataset ID |\n|-|-|\n| Landsat 5 TM Collection 1 Level 1 | `landsat_tm_c1` |\n| Landsat 5 TM Collection 2 Level 1 | `landsat_tm_c2_l1` |\n| Landsat 5 TM Collection 2 Level 2 | `landsat_tm_c2_l2` |\n| Landsat 7 ETM+ Collection 1 Level 1 | `landsat_etm_c1` |\n| Landsat 7 ETM+ Collection 2 Level 1 | `landsat_etm_c2_l1` |\n| Landsat 7 ETM+ Collection 2 Level 2 | `landsat_etm_c2_l2` |\n| Landsat 8 Collection 1 Level 1 | `landsat_8_c1` |\n| Landsat 8 Collection 2 Level 1 | `landsat_ot_c2_l1` |\n| Landsat 8 Collection 2 Level 2 | `landsat_ot_c2_l2` |\n| Sentinel 2A | `sentinel_2a` |\n\n\n# Quick start\n\nSearching for Landsat 5 TM scenes that contains the location (12.53, -1.53) acquired during the year 1995.Set the \n\n```\nusgsm2m search --dataset LANDSAT_TM_C1 --location 30.32 78.03 \\\n    --start 1995-01-01 --end 1995-12-31\n```\n\nSearch for Landsat 7 ETM scenes in Brussels with less than 5% of clouds. Save the returned results in a `.csv` file.\n\n```\nusgsm2m search --dataset LANDSAT_ETM_C1 \\\n    --location 50.83 4.38 --clouds 5 > results.csv\n```\n\nDownloading three Landsat scenes from different datasets in the current directory.\n\n```\nusgsm2m download LT51960471995178MPS00 LC80390222013076EDC00 LC82150682015350LGN01\n```\n\nDownloading three Landsat scenes from different the entity file containing **display id** in ./downloadDir\n\n```\ndownloadbulk --entityfile 143042-cloud-5.csv --output ./downloadDir\n```\nTo use the package, Earth Explorer credentials are required ([registration](https://ers.cr.usgs.gov/register)).\n\n# Installation\n\nThe package can be installed using pip.\n\n```\npip install usgsm2m\n```\n\n# Usage\n\n**usgsm2m** can be used both through its command-line interface and as a Python module.\n\n## Command-line interface\n\n```\nusgsm2m --help\n```\n\n```\nUsage: usgsm2m [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  --help  Show this message and exit.\n\nCommands:\n  download  Download one or several Landsat scenes.\n  search    Search for Landsat scenes.\n```\n\n### Credentials\n\nCredentials for the Earth Explorer portal can be obtained [here](https://ers.cr.usgs.gov/register/).\n\n`--username` and `--password` can be provided as command-line options or as environment variables:\n\n``` shell\nexport USGSM2M_USERNAME=<your_username>\nexport USGSM2M_PASSWORD=<your_password>\n```\n\n### Searching\n\n```\nusgsm2m search --help\n```\n\n```\nUsage: usgsm2m search [OPTIONS]\n\n  Search for Landsat scenes.\n\nOptions:\n  -u, --username TEXT             EarthExplorer username.\n  -p, --password TEXT             EarthExplorer password.\n  -d, --dataset [landsat_tm_c1|landsat_etm_c1|landsat_8_c1|landsat_tm_c2_l1|landsat_tm_c2_l2|landsat_etm_c2_l1|landsat_etm_c2_l2|landsat_ot_c2_l1|landsat_ot_c2_l2|sentinel_2a]\n                                  Landsat data set.\n  -l, --location FLOAT...         Point of interest (latitude, longitude).\n  -b, --bbox FLOAT...             Bounding box (xmin, ymin, xmax, ymax).\n  -c, --clouds INTEGER            Max. cloud cover (1-100).\n  -s, --start TEXT                Start date (YYYY-MM-DD).\n  -e, --end TEXT                  End date (YYYY-MM-DD).\n  -o, --output [entity_id|display_id|json|csv]\n                                  Output format.\n  -m, --limit INTEGER             Max. results returned.\n  --help                          Show this message and exit.\n```\n\n### Downloading \n\n```\nusgsm2m download --help\n```\n\n```\nUsage: usgsm2m download [OPTIONS] [SCENES]...\n\n  Download one or several Landsat scenes.\n\nOptions:\n  -u, --username TEXT    EarthExplorer username.\n  -p, --password TEXT    EarthExplorer password.\n  -d, --dataset TEXT     Dataset.\n  -o, --output PATH      Output directory.\n  -t, --timeout INTEGER  Download timeout in seconds.\n  --skip\n  --help                 Show this message and exit.\n```\n\nIf the `--dataset` is not provided, the dataset is automatically guessed from the scene identifier. Note that only the newer Landsat Product Identifiers contain information related to collection number and processing level. To download Landsat Collection 2 products, use Product IDs or set the `--dataset` option correctly. The download supports the parallel download.\n\n## Bulk Download\n\n```\nusgsm2m download --help\n```\n\n```\nUsage: cli.py downloadbulk [OPTIONS]\n\n  Download one or several scenes.\n\nOptions:\n  -u, --username TEXT    USGS M2M username.\n  -p, --password TEXT    USGS M2M password.\n  -d, --dataset TEXT     Dataset\n  -e, --entityfile PATH  Entity File Name  [required]\n  -o, --output PATH      Output directory.\n  -f, --filetype TEXT    Download File Type {\'bundle\',\'band\'}\n  -i, --idfield TEXT     Entity Id Type {\'displayId\',\'entityId\'}\n  -t, --timeout INTEGER  Download timeout in seconds.\n  --skip\n  --help                 Show this message and exit.\n```\n\nThe --entityfile must contains the scene identifier in each line.If the `--dataset` is not provided, the dataset is automatically guessed from the scene identifier. Note that only the newer Landsat Product Identifiers contain information related to collection number and processing level. \n\n## API\n\n### USGS API\n\n**usgsm2m** provides an interface to the  USGS M2M JSON API. Please refer to the official ([documentation](https://m2m.cr.usgs.gov/api/docs/json/)) for possible request codes and parameters.\n\n#### Basic usage\n\n``` python\nfrom usgsm2m.api import API\n\n# Initialize a new API instance and get an access key\napi = API(username, password)\n\n# Perform a request. Results are returned in a dictionnary\nresponse = api.request(\n    \'<request_endpoint>\',\n    params={\n        "param_1": value_1,\n        "param_2": value_2\n    }\n)\n\n#\xa0Log out\napi.logout()\n```\n\nPlease refer to the official [JSON API Reference](https://m2m.cr.usgs.gov/api/docs/json/) for a list of all available requests.\n\n#### Searching for scenes\n\n``` python\nimport json\nfrom USGSM2M.api import API\n\n# Initialize a new API instance and get an access key\napi = API(username, password)\n\n# Search for Landsat TM scenes\nscenes = api.search(\n    dataset=\'landsat_tm_c1\',\n    latitude=50.85,\n    longitude=-4.35,\n    start_date=\'1995-01-01\',\n    end_date=\'1995-10-01\',\n    max_cloud_cover=10\n)\n\nprint(f"{len(scenes)} scenes found.")\n\n# Process the result\nfor scene in scenes:\n    print(scene[\'acquisition_date\'].strftime(\'%Y-%m-%d\'))\n    # Write scene footprints to disk\n    fname = f"{scene[\'landsat_product_id\']}.geojson"\n    with open(fname, "w") as f:\n        json.dump(scene[\'spatial_coverage\'].__geo_interface__, f)\n\napi.logout()\n```\n\nOutput:\n\n```\n4 scenes found.\n1995-09-23\n1995-08-22\n1995-08-15\n1995-06-28\n```\n\n#### Downloading scenes\n\n``` python\nfrom usgsm2m.usgsm2m import USGSM2M\n\nee = USGSM2M(username, password)\n\nee.download(\'LT51960471995178MPS00\', output_dir=\'./data\')\n\nee.logout()\n```\n',
    'author': 'Ashutosh Kumar Jha',
    'author_email': 'jha.ashutosh@gamil.com',
    'maintainer': 'Ashutosh Kumar Jha',
    'maintainer_email': 'jha.ashutosh@gamil.com',
    'url': 'https://github.com/ashutoshkumarjha/usgsm2m',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Command-line interface."""

import os
import json
import csv
from io import StringIO

import click

from usgsm2m.api import API
from usgsm2m.usgsm2m import USGSM2M
from usgsm2m.errors import USGSM2MError

DATASETS = [
    "landsat_tm_c1",
    "landsat_etm_c1",
    "landsat_8_c1",
    "landsat_tm_c2_l1",
    "landsat_tm_c2_l2",
    "landsat_etm_c2_l1",
    "landsat_etm_c2_l2",
    "landsat_ot_c2_l1",
    "landsat_ot_c2_l2",
    "sentinel_2a",
]


@click.group()
def cli():
    pass


@click.command()
@click.option(
    "-u",
    "--username",
    type=click.STRING,
    help="USGS M2M username.",
    envvar="USGSM2M_USERNAME",
)
@click.option(
    "-p",
    "--password",
    type=click.STRING,
    help="USGS M2M password.",
    envvar="USGSM2M_PASSWORD",
)
@click.option(
    "-d",
    "--dataset",
    type=click.Choice(DATASETS, case_sensitive=False),
    help="Dataset.",
    default="LANDSAT_8_C1",
)
@click.option(
    "-l",
    "--location",
    type=click.FLOAT,
    nargs=2,
    help="Point of interest (latitude, longitude).",
)
@click.option(
    "-b",
    "--bbox",
    type=click.FLOAT,
    nargs=4,
    help="Bounding box (xmin, ymin, xmax, ymax).",
)
@click.option("-c", "--clouds", type=click.INT, help="Max. cloud cover (1-100).")
@click.option("-s", "--start", type=click.STRING, help="Start date (YYYY-MM-DD).")
@click.option("-e", "--end", type=click.STRING, help="End date (YYYY-MM-DD).")
@click.option(
    "-o",
    "--output",
    type=click.Choice(["entity_id", "display_id", "json", "csv"]),
    default="display_id",
    help="Output format.",
)
@click.option("-m", "--limit", type=click.INT, help="Max. results returned.")
def search(
    username, password, dataset, location, bbox, clouds, start, end, output, limit
):
    """Search for scenes."""
    api = API(username, password)

    where = {"dataset": dataset}
    if location:
        latitude, longitude = location
        where.update(latitude=latitude, longitude=longitude)
    if bbox:
        where.update(bbox=bbox)
    if clouds:
        where.update(max_cloud_cover=clouds)
    if start:
        where.update(start_date=start)
    if end:
        where.update(end_date=end)
    if limit:
        where.update(max_results=limit)

    results = api.search(**where)
    api.logout()

    if not results:
        return

    if output == "entity_id":
        for scene in results:
            click.echo(scene["entity_id"])

    if output == "display_id":
        for scene in results:
            click.echo(scene["display_id"])

    if output == "json":
        dump = json.dumps(results, indent=True)
        click.echo(dump)

    if output == "csv":
        with StringIO("tmp.csv") as f:
            w = csv.DictWriter(f, results[0].keys())
            w.writeheader()
            w.writerows(results)
            click.echo(f.getvalue())


@click.command()
@click.option(
    "-u",
    "--username",
    type=click.STRING,
    help="USGS M2M username.",
    envvar="USGSM2M_USERNAME",
)
@click.option(
    "-p",
    "--password",
    type=click.STRING,
    help="USGS M2M password.",
    envvar="USGSM2M_PASSWORD",
)
@click.option("--dataset", "-d", type=click.STRING, required=False, help="Dataset")
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=True, dir_okay=True),
    default=".",
    help="Output directory.",
)
@click.option(
    "--filetype",
    "-f",
    required=False,
    type=click.STRING,
    default="bundle",
    help="Download File Type {'bundle','band'}",
)
@click.option(
    "--idfield",
    "-i",
    required=False,
    type=click.STRING,
    default="displayId",
    help="Entity Id Type {'displayId','entityId'}",
)
@click.option(
    "--timeout", "-t", type=click.INT, default=300, help="Download timeout in seconds."
)
@click.option("--skip", is_flag=True, default=False)
@click.argument("scenes", type=click.STRING, nargs=-1) ############################
def download(username, password, dataset, output, filetype, idfield,timeout, skip, scenes):
    """Download one or several scenes."""
    usgsm2m = USGSM2M(username, password)
    output_dir = os.path.abspath(output)
    if dataset and dataset.lower() not in DATASETS:
        raise USGSM2MError(f"`{dataset}` is not a supported dataset.")
#for scene in scenes:
    if not usgsm2m.logged_in():
        usgsm2m = USGSM2M(username, password)
    #print('cli.py:download ',list(scenes))
    fname = usgsm2m.download(
        scenes, output_dir,filetype=filetype, idField=idfield,dataset=dataset,timeout=timeout, skip=skip
    )

    if skip:
        click.echo(fname)
    usgsm2m.logout()
#######################################################################################

@click.command()
@click.option(
    "--username",
    "-u",
    type=click.STRING,
    help="USGS M2M username.",
    envvar="USGSM2M_USERNAME",
)
@click.option(
    "--password",
    "-p",
    type=click.STRING,
    help="USGS M2M password.",
    envvar="USGSM2M_PASSWORD",
)
@click.option("--dataset", "-d", type=click.STRING, required=False, help="Dataset")
@click.option(
    "--entityfile",
    "-e",
    required=True,
    type=click.Path(exists=True, file_okay=True),
    help="Entity File Name",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(exists=True, dir_okay=True),
    default="./",
    help="Output directory.",
)
@click.option(
    "--filetype",
    "-f",
    required=False,
    type=click.STRING,
    default="bundle",
    help="Download File Type {'bundle','band'}",
)
@click.option(
    "--idfield",
    "-i",
    required=False,
    type=click.STRING,
    default="displayId",
    help="Entity Id Type {'displayId','entityId'}",
)
@click.option(
    "--timeout", "-t", type=click.INT, default=300, help="Download timeout in seconds."
)
@click.option("--skip", is_flag=True, default=False)
def downloadbulk(username, password, entityfile,dataset, output, filetype,idfield,timeout, skip ): #idField
    """Download one or several scenes."""
    usgsm2m = USGSM2M(username, password)
    output_dir = os.path.abspath(output)
    entityfile_path=os.path.abspath(entityfile)
    if dataset and dataset.lower() not in DATASETS:
        raise USGSM2MError(f"`{dataset}` is not a supported dataset.")
    if not usgsm2m.logged_in():
        usgsm2m = USGSM2M(username, password)
    fname = usgsm2m.downloadbulk(
        entityfile=entityfile_path, output_dir=output_dir, filetype=filetype,dataset=dataset, idField=idfield,timeout=timeout, skip=skip
    )
    if skip:
        click.echo(fname)
    usgsm2m.logout()


cli.add_command(search)
cli.add_command(download)
cli.add_command(downloadbulk)


if __name__ == "__main__":
    cli()
